/* Suchalgorithmus nach Knuth, Morris, Pratt. */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>



int main(int argc, char** argv)
{
    if (argc < 3)
    {
        printf("Usage:\n\tkmp needle haystack\n\n");
        return 1;
    }

    char* needle = argv[1];
    int maxNeedle = strlen(needle);
    int* pattern = (int*) malloc(maxNeedle * sizeof(int));


    /* Pattern der needle ermitteln. */

    int posNeedle = 1, cntMatch = 0;

    for (posNeedle = 1; posNeedle < maxNeedle; posNeedle++)
    {
        if (needle[posNeedle] == needle[cntMatch])
        {
            cntMatch++;
            pattern[posNeedle] = cntMatch;
        }
        else
        {
            /* Zuruecksetzen an den Anfang. */
            cntMatch = 0;
            pattern[posNeedle] = cntMatch;

            if (needle[posNeedle] == needle[cntMatch])
            {
                /* Pattern in needle beginnt erneut von vorn. */
                cntMatch++;
                pattern[posNeedle] = cntMatch;
            }
            else
            {
                /* Pattern in needle beginnt nicht erneut von vorn.
                 * Zuruecksetzung bereits erfolgt. */
            }
        }
    }


    int posHaystack = 0;
    char* haystack = argv[2];
    int maxHaystack = strlen(haystack);

    for (posHaystack = 0, posNeedle = 0, cntMatch = 0;
         /* Optimierung fuer posHaystack < maxHaystack. */
         (posHaystack - cntMatch) + (maxNeedle - 1) < maxHaystack;
         posHaystack++)
    {
        printf("%d %d\n", posHaystack, posNeedle);

        if (haystack[posHaystack] == needle[posNeedle])
        {
            posNeedle++;
            cntMatch++;

            if (cntMatch == maxNeedle)
            {
                /* cntMatch - 1, da Positionen nullbasierend. */
                printf("Found at %d.\n", posHaystack - (cntMatch - 1));

#if 1
/* Optimierung fuer den else-Teil. */

                posHaystack -= pattern[maxNeedle - 1];
                /* posHaystack -= 1 unterbleibt wegen posHaystack++ zu Beginn
                 * der naechsten Schleifen-Iteration. */

                printf("Jump saved %d character checks.\n", pattern[maxNeedle - 1]);

                posNeedle = 0;
                cntMatch = 0;
#else
                /* Ab dem zweiten Buchstaben des gefundenen needle
                 * neue Suche starten. */
                posHaystack = posHaystack - (cntMatch - 1);
                posNeedle = 0;
                cntMatch = 0;
#endif
            }
        }
        else
        {
            posNeedle = 0;

            if (cntMatch > 0)
            {
                printf("Jump saved %d character checks.\n", pattern[cntMatch - 1]);

                posHaystack -= pattern[cntMatch - 1];
                /* -1 wegen posHaystack++ zu Beginn der naechsten
                 * Schleifen-Iteration. */
                posHaystack -= 1;
                cntMatch = 0;
            }
        }
    }

    free(pattern);
    pattern = NULL;

    return 0;
}

